{: Main form of the demo showing how to use worker threads.}
unit ThreadDemoFrm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WorkUnit, PrimeWork, StdCtrls, Spin, ActnList;

type
  {: Main form for the worker thread demonstration app.}
  TThreadDemoForm = class(TForm)
    GoButton: TButton;
    ActionList1: TActionList;
    ThreadSpin: TSpinEdit;
    Label1: TLabel;
    WorkListBox: TListBox;
    Label2: TLabel;
    Label3: TLabel;
    CompletedListBox: TListBox;
    Label4: TLabel;
    MaxSpin: TSpinEdit;
    Label5: TLabel;
    NumberSpin: TSpinEdit;
    GoAction: TAction;
    Label6: TLabel;
    ThreadsListbox: TListBox;
    Button1: TButton;
    OpenDialog: TOpenDialog;
    procedure GoExecute(Sender: TObject);
    procedure QueueWork ;
    procedure CreateWork( const AMaxPrime, ANumberPerItem : integer ) ;
    procedure CreateThreads( const ANumber : integer ) ;
    procedure GoActionUpdate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    FTerminationEvent : THandle ;
    FWorkItems        : TList ;
    FWorkList         : TWorkList ;
    FCompletionList   : TWorkList ;
    FThreads          : array of TWorkerThread ;
    procedure FreeListItems( const AList : TList ) ;
    procedure FreeThreads ;
    procedure SetThreadCount(const Value: integer);
    procedure wmWorkComplete( var Msg : TMessage ) ; message WM_WORKCOMPLETE ;
    function  GetIdleString( const AIndex : integer ) : string ;
    procedure WorkHandler( const Sender      : TWorkerThread ;
                           const AWork       : TWork ;
                           const AWorkStatus : TWorkStatus ) ;
  public
    constructor Create( AOwner : TComponent ) ; override ;
    destructor Destroy ; override ;
    property  ThreadCount : integer write SetThreadCount ;
  end;

var
  ThreadDemoForm: TThreadDemoForm;

implementation

{$R *.DFM}

{ TThreadDemoForm }

{: Creates a manual reset event for termination of the threads, a list to own work items
   and lists for the work as it is done.}
constructor TThreadDemoForm.Create(AOwner: TComponent);
begin
  inherited;
  FTerminationEvent := CreateEvent( NIL, true, false, NIL ) ;
  FWorkItems := TList.Create ;
  FWorkList := TWorkList.Create( 5 ) ;
  FCompletionList := TWorkList.Create( 5 ) ;
end;

{: Add the work in the work items list to the work list and the listbox.}
procedure TThreadDemoForm.QueueWork;
var i : integer ;
    AWork : TWork ;
begin
  with WorkListbox.Items do begin
    BeginUpdate ;
    try
      Clear ;
      for i := 0 to FWorkItems.Count - 1 do begin
        AWork := FWorkItems[ i ] ;
        FWorkList.Add( AWork ) ;
        AddObject( AWork.Identification, AWork ) ;
      end ;
    finally
      EndUpdate ;
    end ;
  end ;
end;

{: Frees the threads, waiting for them as necessary, then tidies up by
   freeing the work.}
procedure TThreadDemoForm.CreateThreads( const ANumber : integer );
var i : integer ;
    AThread : TWorkerThread ;
begin
  // free any existing threads now
  FreeThreads ;

  // now add 'em
  SetLength( FThreads, ANumber ) ;
  FillChar( FThreads[ 0 ], ANumber * SizeOf( pointer ), 0 ) ;
  for i := 0 to ANumber - 1 do begin
    AThread := TWorkerThread.Create( false, FWorkList, FCompletionList,
                                           Self, FTerminationEvent ) ;
    try
      AThread.OnWork := WorkHandler ;
      FThreads[ i ] := AThread ;
      ThreadsListbox.Items.AddObject( GetIdleString( i ), AThread ) ;

      // reduce the thread priority so that the foreground thread remains
      // the most responsive
      AThread.Priority := tpLower ;
    except
      AThread.Free ;
      Raise ;
    end ;
  end ;
end;

{: Populate the work items list with items of work.}
procedure TThreadDemoForm.CreateWork( const AMaxPrime, ANumberPerItem : integer );
var i, ACount : integer ;
    AWork : TWork ;
begin
  FreeListItems( FWorkItems ) ;
  i := 1 ;
  ACount := 0 ;
  while i <= AMaxPrime do begin
    inc( ACount ) ;
    AWork := TPrimeWork.Create( Format( 'Item %d, %d to %d', [ ACount, i, i + ANumberPerItem - 1 ] ),
                                i, i + ANumberPerItem - 1 ) ;
    try
      FWorkItems.Add( AWork ) ;
    except
      AWork.Free ;
      Raise ;
    end ;
    inc( i, ANumberPerItem ) ;
  end ;
end;

destructor TThreadDemoForm.Destroy;
begin
  FreeThreads ;
  FreeListItems( FWorkItems ) ;
  FWorkList.Free ;
  FCompletionList.Free ;
  FWorkItems.Free ;
  CloseHandle( FTerminationEvent ) ;
  inherited;
end;

{: Frees all the items in the list and clears it.}
procedure TThreadDemoForm.FreeListItems(const AList: TList);
var i : integer ;
begin
  for i := 0 to AList.Count - 1 do
    TObject( AList[ i ] ).Free ;
  AList.Clear ;
end;

{: Sets the termination event, then frees the threads and resets the event.}
procedure TThreadDemoForm.FreeThreads;
var i : integer ;
begin
  // clear the listbox
  ThreadsListbox.Clear ;

  // signal the termination event to tell all the threads to stop
  SetEvent( FTerminationEvent ) ;
  try
    // now free them all
    for i := Low( FThreads ) to High( FThreads ) do
      with FThreads[ i ] do
    begin
      // increase thread priority to speed termination
      Priority := tpHigher ;
      Free ;
    end ;
    SetLength( FThreads, 0 ) ;
  finally
    ResetEvent( FTerminationEvent ) ;
  end ;
end;

{: Create the work for the threads to do.}
procedure TThreadDemoForm.GoExecute(Sender: TObject);
begin
  CompletedListbox.Clear ;
  ThreadCount := ThreadSpin.Value ;
  CreateWork( MaxSpin.Value, NumberSpin.Value ) ;
  QueueWork ;
end;

{: Creates threads if the number requested differs from the current number.}
procedure TThreadDemoForm.SetThreadCount(const Value: integer);
begin
  // if we already have this number of threads, do nothing
  if Length( FThreads ) = Value then exit ;
  FreeThreads ;
  CreateThreads( Value ) ;
end;

{: Responds to the WM_WORKCOMPLETE message posted by <see class=TWorkerThread>
   when work has been performed.}
procedure TThreadDemoForm.wmWorkComplete(var Msg: TMessage);
var AWork  : TPrimeWork ;
    AIndex : integer ;
begin
  // remove an item from the complete list
  AWork := FCompletionList[ 0 ] as TPrimeWork ;
  FCompletionList.Delete( 0 ) ;

  // remove it from the work list box
  AIndex := WorkListbox.Items.IndexOfObject( AWork ) ;
  if AIndex >= 0 then
    WorkListbox.Items.Delete( AIndex ) ;

  // add it to the completed list box
  CompletedListbox.Items.AddObject( Format( '%s - %d primes', [ AWork.Identification, AWork.Results.Count ] ),
                                    AWork ) ;
end;

{: The "go" action is only enabled when there is no work queued.}
procedure TThreadDemoForm.GoActionUpdate(Sender: TObject);
begin
  ( Sender as TAction ).Enabled := WorkListbox.Items.Count = 0 ;
end;

{: Handles the <see property=OnWork class=TWorkerThread Text=OnWork> event of
   the worker threads by showing that the thread is performing the given work
   item, or is idle.}
procedure TThreadDemoForm.WorkHandler(const Sender: TWorkerThread;
  const AWork: TWork; const AWorkStatus: TWorkStatus);
var AIndex : integer ;
    s : string ;
begin
  // find the thread in the listbox
  AIndex := ThreadsListbox.Items.IndexOfObject( Sender ) ;
  if AIndex < 0 then exit ;

  // update the listbox
  if AWorkStatus = wsStart then
    s := Format( 'Thread %d working on %s', [ AIndex + 1, AWork.Identification ] ) else
    s := GetIdleString( AIndex ) ;
  ThreadsListbox.Items[ AIndex ] := s ;
  ThreadsListbox.Items.Objects[ AIndex ] := Sender ; // just to make sure
end;

{: Returns a simple string showing that a given thread is idle.}
function TThreadDemoForm.GetIdleString(
  const AIndex : integer): string;
begin
  Result := Format( 'Thread %d idle', [ AIndex + 1 ] ) ;
end;

{: Demonstrates that the application is responsive simply by opening a browse dialog
   and reporting any file chosen.}
procedure TThreadDemoForm.Button1Click(Sender: TObject);
begin
  if OpenDialog.Execute then
    MessageDlg( Format( 'You picked %s', [ OpenDialog.FileName ] ),
                mtInformation, [ mbOK ], 0 ) ;
end;

end.
